#!/usr/bin/env python

"""
Common classes.

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import compiler.visitor
import operator

try:
    set
except NameError:
    from sets import Set as set

class ObjectSet:

    "A set of objects with optional associated data."

    def __init__(self, d=None):
        if d is None:
            self.objects = {}
        elif hasattr(d, "items"):
            self.objects = dict(d)
        else:
            self.objects = {}
            for key in d:
                self.add(key)

    def __repr__(self):
        out = ["{"]
        first = 1
        for key, value in self.items():
            if not first:
                out.append(", ")
            else:
                first = 0
            out.append(repr(key))
            if value:
                out.append(" : ")
                out.append(repr(value))
        out.append("}")
        return "".join(out)

    def __iter__(self):
        return iter(self.keys())

    def __nonzero__(self):
        return self.objects != {}

    # List methods.

    def __add__(self, other):
        obj = ObjectSet(self)
        for key in other:
            obj.add(key)
        return obj

    # Set membership and comparisons.

    def __hash__(self):
        return hash(tuple(self.keys()))

    def __eq__(self, other):
        if hasattr(other, "objects"):
            return self.objects == other.objects
        else:
            return set(self.objects.keys()) == set(other)

    # Set methods.

    def add(self, obj):
        if not self.has_key(obj):
            self[obj] = set()

    # Dictionary and related methods.

    def __getitem__(self, key):
        return self.objects[key]

    def get(self, key, default=None):
        return self.objects.get(key, default)

    def __setitem__(self, key, value):
        self.objects[key] = value

    def has_key(self, key):
        return self.objects.has_key(key)

    def keys(self):
        return self.objects.keys()

    def values(self):
        return self.objects.values()

    def items(self):
        return self.objects.items()

    def update(self, other):

        # Combining dictionary-like objects involves combining values.

        if hasattr(other, "items"):
            for key, value in other.items():
                self[key] = add_sets(value, self.get(key, []))

        # Combining sequence-like objects involves just adding members.

        else:
            for key in other:
                self.add(key)

    def merge(self, other):

        """
        Merge this object set with an 'other' set, combining the values where
        possible, and incorporating values present in only one of the sets.
        """

        return combine(self, other, ObjectSet(), add_sets)

def deepen_mapping_dict(d):

    "Convert the values of 'd' to be elements of a potentially larger set."

    new_dict = {}
    for key, value in d.items():
        if value is None:
            new_dict[key] = None
        else:
            new_dict[key] = ObjectSet([value])
    return new_dict

def merge_mapping_dicts(dicts):

    "Merge the given 'dicts' mapping keys to sets of objects."

    new_dict = {}
    update_mapping_dict(new_dict, dicts)
    return new_dict

def update_mapping_dict(new_dict, dicts):

    """
    Update 'new_dict' with the contents of the set dictionary 'dicts'.
    None entries may be incorporated into object sets. For example:

    d1: {'a' : None}
    d2: {'a' : x}
    ->  {'a' : {None, x}}

    Here, None might be used to represent an empty set and x may be an existing
    set.
    """

    for old_dict in dicts:
        for key, value in old_dict.items():
            if not new_dict.has_key(key):
                if value is not None:
                    new_dict[key] = ObjectSet(value)
                else:
                    new_dict[key] = None
            elif new_dict[key] is not None:
                if value is not None:
                    new_dict[key].update(value)
                else:
                    new_dict[key].add(value)
            else:
                if value is not None:
                    objset = new_dict[key] = ObjectSet(value)
                    objset.add(None)

def combine_mapping_dicts(d1, d2):

    """
    Combine dictionaries 'd1' and 'd2' in a resulting dictionary, with the
    values of the contributing dictionaries being themselves combined such that
    a "product" of the values for a given key are stored in the combined
    dictionary.

    For example:

    d1: {'a' : [{'f', 'g'}, {'f', 'h'}], ...}
    d2: {'a' : [{'f'}, {'e', 'f', 'g'}], ...}
    ->  {'a' : [{'f', 'g'}, {'f', 'h'}, {'e', 'f', 'g'}, {'e', 'f', 'g', 'h'}], ...}

    Note that items of 'd2' whose keys are not in 'd1' are not added to 'd1'
    since this, in the context of propagating attribute usage observations,
    would result in spurious usage details being made available in places where
    the names may not have been defined.
    """

    return combine(d1, d2, {}, combine_object_set_lists, True)

def combine(d1, d2, combined, combine_op, only_d1_keys=False):

    """
    Combine dictionaries 'd1' and 'd2' in the 'combined' object provided, using
    the 'combine_op' to merge values from the dictionaries.

    If 'only_d1_keys' is set to a true value, items from 'd2' employing keys not
    in 'd1' will not be added to 'd1'.
    """

    if d2 is not None:
        d2_keys = d2.keys()

        for key in d2_keys:
            if not d1.has_key(key):
                if not only_d1_keys:
                    combined[key] = d2[key]
            else:
                combined[key] = combine_op(d1[key], d2[key])
    else:
        d2_keys = ()

    for key in d1.keys():
        if key not in d2_keys:
            combined[key] = d1[key]

    return combined

def combine_object_set_lists(l1, l2):

    """
    Combine lists of object sets 'l1' and 'l2' to make a product of their
    members.
    """

    # If either list is undefined (indicated by None), return the defined list,
    # or return None if neither is defined.

    if l1 is None:
        if l2 is None:
            return None
        else:
            return l2 + []
    elif l2 is None:
        return l1 + []

    combined = ObjectSet()
    for i1 in l1:
        for i2 in l2:
            combined.add(i1.merge(i2))
    return combined

def add_sets(s1, s2):
    return set(list(s1) + list(s2))

# Visitors and activities related to node annotations.

class ASTVisitor(compiler.visitor.ASTVisitor):

    "A base class for visitors."

    def dispatch(self, node, *args):

        "Dispatch using 'node', annotating any raised exceptions."

        try:
            return compiler.visitor.ASTVisitor.dispatch(self, node, *args)
        except NodeProcessingError, exc:
            if exc.astnode is None:
                exc.astnode = node

                # NOTE: Should perhaps specialise the subclasses appropriately.

                if hasattr(self, "unit"):
                    exc.unit_name = self.unit.full_name()
                else:
                    exc.unit_name = self.full_name()
            raise

    def possible_accessor_types(self, node, defining_users=1):

        """
        Given annotations made during the inspection process, return all possible
        type names and indications of static usage for a 'node' involved in
        attribute access.

        If 'defining_users' is set to a false value, attempt to get the type
        names specifically applicable to the node, rather than retrieving more
        general definition-based type observations.
        """

        target_names = set()

        if hasattr(node, "_attrusers"):

            # Visit each attribute user.

            for user in node._attrusers:

                # Since users such as branches may not provide type information,
                # attempt to find defining users.

                if defining_users:
                    for def_user in user._attrdefs or [user]:
                        for target_name, is_static in def_user._attrtypes.get(node._username, []):
                            target_names.add((target_name, is_static))
                else:
                    for target_name, is_static in user._attrspecifictypes.get(node._username, []):
                        target_names.add((target_name, is_static))

        return target_names

def used_by_unit(node):

    """
    Return whether the definition made by a 'node' is actually employed by the
    program unit within which it is found.
    """

    return hasattr(node, "unit") and node.unit.parent.has_key(node.unit.name)

def get_object_types_for_usage(usage, objtable, name, unit_name):

    """
    Return for the given attribute 'usage', using the 'objtable', the object
    types which satisfy such usage, reporting any problems for the given 'name'
    and 'unit_name'. Each object type is given as a tuple of the form (type,
    is_static).
    """

    all_objtypes = set()

    for attrnames in usage:
        attrnames = attrnames or ()
        objtypes = objtable.all_possible_objects_plus_status(attrnames)
        if not objtypes:
            print "Warning: usage in %r for %r finds no object supporting all attributes %r" % (
                unit_name, name, attrnames)
            objtypes = objtable.any_possible_objects_plus_status(attrnames)
            if not objtypes:
                print "Warning: usage in %r for %r finds no object supporting any attributes %r" % (
                    unit_name, name, attrnames)

        all_objtypes.update(objtypes)

    return all_objtypes

# Errors.

class ProcessingError(Exception):

    "A processing error."

    pass

class TableError(ProcessingError):

    "An error occurring during access to a lookup table."

    pass

class TableGenerationError(ProcessingError):

    "An error occurring when generating a lookup table."

    pass

class NodeProcessingError(ProcessingError):

    "A processing error associated with a particular program node."

    def __init__(self, message):
        self.message = message
        self.unit_name = None
        self.astnode = None

    def get_lineno(self, node):
        if node is None:
            return None

        lineno = node.lineno
        if lineno is not None:
            return lineno
        else:
            for child in node.getChildNodes():
                lineno = self.get_lineno(child)
                if lineno is not None:
                    return lineno
        return None

    def __repr__(self):
        return "Error in %r at line %r: %s" % (self.unit_name, self.get_lineno(self.astnode), self.message)

    def __str__(self):
        return repr(self)

class InspectError(NodeProcessingError):

    "An error during the module inspection process."

    pass

class TranslateError(NodeProcessingError):

    "An error during the module translation process."

    pass

class TranslationNotImplementedError(TranslateError):

    "An error caused by a node not being supported in translation."

    pass

# Special representations.

class AtLeast:

    "A special representation for numbers of a given value or greater."

    def __init__(self, count):
        self.count = count

    def __eq__(self, other):
        return 0

    __lt__ = __le__ = __eq__

    def __ne__(self, other):
        return 1

    def __gt__(self, other):
        if isinstance(other, AtLeast):
            return 0
        else:
            return self.count > other

    def __ge__(self, other):
        if isinstance(other, AtLeast):
            return 0
        else:
            return self.count >= other

    def __iadd__(self, other):
        if isinstance(other, AtLeast):
            self.count += other.count
        else:
            self.count += other
        return self

    def __radd__(self, other):
        if isinstance(other, AtLeast):
            return AtLeast(self.count + other.count)
        else:
            return AtLeast(self.count + other)

    def __repr__(self):
        return "AtLeast(%r)" % self.count

class Location:

    """
    A special representation for locations which are to be compared to program
    objects.
    """

    def __init__(self, location):
        self.location = location

    def _op(self, other, op):
        if hasattr(other, "location"):
            return op(self.location, other.location)
        else:
            raise NotImplemented

    def __eq__(self, other):
        return self._op(other, operator.eq)

    def __ne__(self, other):
        return self._op(other, operator.ne)

    def __lt__(self, other):
        return self._op(other, operator.lt)

    def __le__(self, other):
        return self._op(other, operator.le)

    def __gt__(self, other):
        return self._op(other, operator.gt)

    def __ge__(self, other):
        return self._op(other, operator.ge)

    def __repr__(self):
        return "Location(%r)" % self.location

# Useful data.

operator_functions = {

    # Binary operations.

    "Add" : "add",
    "Bitand" : "and_",
    "Bitor" : "or_",
    "Bitxor" : "xor",
    "Div" : "div",
    "FloorDiv" : "floordiv",
    "LeftShift" : "lshift",
    "Mod" : "mod",
    "Mul" : "mul",
    "Power" : "pow",
    "RightShift" : "rshift",
    "Sub" : "sub",

    # Unary operations.

    "Invert" : "invert",
    "UnaryAdd" : "pos",
    "UnarySub" : "neg",

    # Augmented assignment.

    "+=" : "iadd",
    "-=" : "isub",
    "*=" : "imul",
    "/=" : "idiv",
    "//=" : "ifloordiv",
    "%=" : "imod",
    "**=" : "ipow",
    "<<=" : "ilshift",
    ">>=" : "irshift",
    "&=" : "iand",
    "^=" : "ixor",
    "|=" : "ior",

    # Comparisons.

    "==" : "eq",
    "!=" : "ne",
    "<" : "lt",
    "<=" : "le",
    ">=" : "ge",
    ">" : "gt",

    # Access and slicing.

    "AssSlice" : "setslice",
    "Slice" : "getslice",
    "AssSubscript" : "setitem",
    "Subscript" : "getitem",
    }

# vim: tabstop=4 expandtab shiftwidth=4
